@@ -6,7 +6,7 @@
   $Date: 1995/01/10 10:42:38 $
   created at: Tue Dec 28 16:01:58 JST 1993
 
-  Copyright (C) 1993-1995 Yukihiro Matsumoto
+  Copyright (C) 1993-1996 Yukihiro Matsumoto
 
 ************************************************/
 
@@ -18,16 +18,20 @@ rb_call_inits()
     Init_sym();
     Init_var_tables();
     Init_Object();
+#ifdef THREAD
+    Init_Thread();
+#endif
     Init_GC();
     Init_eval();
     Init_Comparable();
     Init_Enumerable();
+    Init_String();
+    Init_Exception();
     Init_Numeric();
     Init_Bignum();
     Init_Array();
     Init_Hash();
     Init_Struct();
-    Init_String();
     Init_Regexp();
     Init_pack();
     Init_Range();
